﻿// BenLincoln.TheLostWorlds.CDHashes
// Copyright 2018 Ben Lincoln
// https://www.thelostworlds.net/
//

// This file is part of BenLincoln.TheLostWorlds.CDHashes.

// BenLincoln.TheLostWorlds.CDHashes is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// BenLincoln.TheLostWorlds.CDHashes is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with BenLincoln.TheLostWorlds.CDHashes (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

/* Hashing algorithm used in all versions of Blood Omen */

namespace BenLincoln.TheLostWorlds.CDHashes
{
    public class BloodOmenHash : CDHash
    {
        public BloodOmenHash()
            : base()
        {
        }

        public override byte[] GetHashBytes(string input, bool hashRawText)
        {
            string inVal = input;

            char[] inChars;
            uint nameHash;
            int lengthDiff;
            string inValModified = inVal;
            // convert to uppercase, replace slashes with colons
            /*inValModified = inValModified.ToUpper().Trim();
            inValModified = inValModified.Replace('/', ':');
            inValModified = inValModified.Replace('\\', ':');*/

            inChars = inValModified.ToCharArray();

            lengthDiff = inChars.Length % 4;

            // Blood Omen's hash algorithm operates in groups of four characters
            // so if the filename is not four characters long, it needs to be padded out
            if (lengthDiff > 0)
            {
                char[] tempChars = new char[inChars.Length + (4 - lengthDiff)];
                for (int i = 0; i < tempChars.Length; i++)
                {
                    tempChars[i] = (char)0;
                }
                for (int i = 0; i < inChars.Length; i++)
                {
                    tempChars[i] = inChars[i];
                }
                inChars = tempChars;
            }

            nameHash = 0;
            for (int baseCharNum = 0; baseCharNum < inChars.Length; baseCharNum += 4)
            {
                uint currentGroup = 0;
                currentGroup = (((uint)inChars[baseCharNum + 3]) << 24) | (((uint)inChars[baseCharNum + 2]) << 16) | (((uint)inChars[baseCharNum + 1]) << 8) |
                    (inChars[baseCharNum]);
                nameHash = nameHash ^ currentGroup;
            }

            return UIntToByteArray(nameHash);
        }
    }
}
