﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;

namespace BenLincoln.TheLostWorlds.CDHashes
{
    public class SoulReaverDreamcastHash : CDHash
    {
        public SoulReaverDreamcastHash()
            : base()
        {
        }

        // From the SH4 docs:
        // for shld Rm,Rn:
        // If Rm >= 0: Rn << Rm -> Rn
        // If Rm < 0: Rn >> |Rm| -> [0 -> Rn]
        public static int sh4shld(int val1, int val2)
        {
            int result = 0;
            if (val1 >= 0)
            {
                result = val2 << val1;
            }
            else
            {
                result = val2 >> Math.Abs(val1);
            }
            return result;
        }

        [StructLayout(LayoutKind.Explicit)]
        struct UIntIntOverlay
        {
            [FieldOffset(0)]
            public uint UInt;
            [FieldOffset(0)]
            public int Int;
        }

        public static uint sh4shld(uint val1, uint val2)
        {
            UIntIntOverlay overlayVal1;
            UIntIntOverlay overlayVal2;
            UIntIntOverlay overlayResult;
            overlayVal1.Int = 0;
            overlayVal2.Int = 0;
            overlayResult.UInt = 0;
            overlayVal1.UInt = val1;
            overlayVal2.UInt = val2;
            overlayResult.Int = sh4shld(overlayVal1.Int, overlayVal2.Int);
            return overlayResult.UInt;
        }

        public override byte[] GetHashBytes(string input, bool hashRawText)
        {
            string inVal = input;

            uint charsRead = 0;
            int currentLetter = 0;
            int index = 0;
            int extID = 0;
            uint hashName = 0;
            uint hashName2 = 0;
            uint hashName3 = 0;

            uint r0 = 0;
            uint r1 = 0;
            uint r4 = 0;
            uint r6 = 0;
            uint r7 = 0;
            // r12 == charsRead
            uint r14 = 0;
            uint r15 = 0;
            uint macl = 0;

            if (hashRawText)
            {
                index = inVal.Length - 1;
            }
            else
            {
                // in the Dreamcast code, the "\kain2\" at the beginning of paths is stripped
                if (inVal.Substring(0, 7).ToLower() == @"\kain2\")
                {
                    inVal = inVal.Substring(7);
                }
                if (inVal.Contains("."))
                {
                    string extension = inVal.Substring(inVal.LastIndexOf('.') + 1).ToLower();
                    string[] extensions = { "pcm", "crm", "tim", "pmp", "pnd", "pmf", "pnf" };
                    for (index = 0; index < extensions.Length; index++)
                    {
                        if (extension.Equals(extensions[index]))
                        {
                            hashName3 = (uint)index;
                            break;
                        }
                    }
                    // if the extension is known, remove it and the period
                    if (index < 7)
                    {
                        index = inVal.Length - 5;
                    }
                    // otherwise, hash everything, including the extension
                    else
                    {
                        index = inVal.Length - 1;
                    }
                }
                else
                {
                    // just process the entire string
                    index = inVal.Length - 1;
                }
            }

            uint r9 = (uint)index;

            while (index >= 0)
            {
                currentLetter = (int)inVal[index];
                r0 = (uint)currentLetter;
                if (currentLetter >= 'a' && currentLetter <= 'z')
                {
                    r0 = (uint)currentLetter & 0xDF;
                }
                if (currentLetter == 0x5c)
                {
                    index--;
                    r9--;
                    continue;
                }
                
                r0 -= 0x1A;
                r0 &= 0xFF;
                r1 = r0 * r9;
                macl = r0 * charsRead;
                r7 = hashName;
                r6 = hashName2;
                r7 += r1;
                r1 = macl;
                r6 ^= r1;
                hashName = r7;
                charsRead++;
                hashName2 = r6;
                index--;
                r9--;
            }
            r7 = 0x1B;
            r6 = hashName;
            uint r12 = sh4shld(r7, charsRead);
            r7 = 0x0F;
            r6 = sh4shld(r7, r6);
            r12 |= r6;
            r6 = hashName2;
            r6 = r6 << 2;
            r7 = hashName3;
            r6 *= 2;
            r12 |= r6;
            r12 |= r7;
            r14 += 0x0C;
            r0 = r12;
            r15 = r14;

            return UIntToByteArray(r12);
        }
    }
}
